// Copyright 2019 PDFium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

// Original code copyright 2014 Foxit Software Inc. http://www.foxitsoftware.com

#include "core/fxge/text_char_pos.h"

TextCharPos::TextCharPos() = default;

TextCharPos::TextCharPos(const TextCharPos&) = default;

TextCharPos::~TextCharPos() = default;
